import ast
import moncov
import collections
import os
from rate import Rate

class Visitor(ast.NodeVisitor):
    def __init__(self, hit_count={}):
        self.hit_count = hit_count
        self.lines = set()
        self.branches = set()
        self.conditions = {}
        self.line_rate = None
        self.branch_rate = None

    def add_line(self, node):
        '''ad node and update self.line_rate return True if line was added'''
        if not hasattr(node, 'lineno') or node.lineno in self.lines:
            # already encountered --- skip
            return False
        self.lines.add(node.lineno)
        if node.lineno in self.hit_count:
            # this line was executed
            rate = Rate(1, 1)
        else:
            rate = Rate(0, 1)
        if self.line_rate is None:
            # initial state
            self.line_rate = rate
        else:
            self.line_rate |= rate
        return True

    def get_condition_rate(self, node):
        '''figure out current node's condition rate/coverate'''
        orelse_hits = set([subnode.lineno for subnode in node.orelse if
                            hasattr(subnode, 'lineno') and subnode.lineno in self.hit_count])
        body_hits = set([subnode.lineno for subnode in node.body if
                            hasattr(subnode, 'lineno') and subnode.lineno in self.hit_count])
        if node.orelse:
            # node has both orelse and body branch

            if orelse_hits and body_hits:
                # 2 branches and 2 hits
                rate = Rate(2, 2)
            elif orelse_hits or body_hits:
                # either orelse or body without hits
                rate = Rate(1, 2)
            else:
                # no hits at all
                rate = Rate(0, 2)
        else:
            # just body branch
            if body_hits:
                # body was executed
                rate = Rate(1, 1)
            else:
                rate = Rate(0, 1)
        return rate

    def get_branch_rate(self, node):
        '''figure out current node's branch rate'''
        rate = self.get_condition_rate(node)
        for subnode in node.body + node.orelse:
            sub_rate = self.visit(subnode)
            if sub_rate is not None:
                sub_rate |= rate
        return rate


    def add_branch(self, node):
        '''add a branched node and update self.branch_rate'''
        branch_rate = self.get_branch_rate(node)
        if self.branch_rate is None:
            # initial state
            self.branch_rate = branch_rate
        else:
            self.branch_rate |= branch_rate
        self.branches.add(node.lineno)
        self.conditions[node.lineno] = self.get_condition_rate(node)

    def visit_If(self, node):
        '''visit all If-sub nodes and update branch rate'''
        self.add_line(node)
        self.visit(node.test)
        return self.add_branch(node)

    def visit_While(self, node):
        self.add_line(node)
        self.visit(node.test)
        return self.add_branch(node)

    def visit_For(self, node):
        self.add_line(node)
        self.visit(node.target)
        self.visit(node.iter)
        return self.add_branch(node)

    def generic_visit(self, node):
        self.add_line(node)
        return super(Visitor, self).generic_visit(node)


FileStatus = collections.namedtuple('FileStatus', ['filename',
                'branch_rate', 'line_rate', 'branches', 'lines', 'conditions', 'hit_count'])
FileErrorStatus = collections.namedtuple('FileErrorStatus', ['filename', 'error'])

def get_stats(db=None, whitelist=None, blacklist=None):
    if db is None:
        db = moncov.conf.get_db()
    whitelist = moncov.conf.get_whitelist(whitelist)
    blacklist = moncov.conf.get_blacklist(blacklist)

    stats = []

    for filename in moncov.data.filenames(db):
        # does the filename require processing?
        if not any([pattern.match(filename) for pattern in whitelist]) or \
            any([pattern.match(filename) for pattern in blacklist]):
            stats.append(FileErrorStatus(filename=filename, error=ValueError('filtered-away')))
            continue
        # fetch the filename content
        try:
            with open(filename) as fd:
                src = fd.read()
        except Exception as e:
            # error reading filename --- mark an error status
            stats.append(FileErrorStatus(filename=filename, error=e))
            continue
        # parse source filename content
        try:
            tree = ast.parse(src)
        except Exception as e:
            # not valid python code --- could happen because file changed
            # between being traced and processed here
            stats.append(FileErrorStatus(filename=filename, error=e))
            continue
        # fetch lineno--hitcount stats
        hit_count = {
            moncov.data.arc2line(arc): moncov.data.filename_arc_hits(db, filename, arc) \
                for arc in moncov.data.filename_arcs(db, filename)
        }
        # calculate the source file AST rates
        visitor = Visitor(hit_count=hit_count)
        visitor.visit(tree)
        # record the source filename stats
        stats.append(FileStatus(filename=filename, line_rate=visitor.line_rate,
                branch_rate=visitor.branch_rate, lines=visitor.lines, branches=visitor.branches,
                hit_count=visitor.hit_count, conditions=visitor.conditions))

    return stats

def print_stats(db=None, whitelist=None, blacklist=None):
    stats = get_stats(db=db, whitelist=whitelist, blacklist=blacklist)
    print "# filename, linerate, branchrate"
    for file_status in stats:
        if isinstance(file_status, FileStatus):
            if file_status.branch_rate is not None:
                branch_rate = '%1.2f' % file_status.branch_rate
            else:
                branch_rate = 'N/A'
            if file_status.line_rate is not None:
                line_rate = '%1.2f' % file_status.line_rate
            else:
                line_rate = 'N/A'
            print  file_status.filename + ', ' + line_rate + ', ' + branch_rate
        elif isinstance(file_status, FileErrorStatus):
            print "# %s: %r" % (file_status.filename, file_status.error)
        else:
            print "# bad file_status: %r" % file_status

def get_stats_xml(db=None, whitelist=None, blacklist=None):
        '''return a header-only cobertura xml'''
        stats = get_stats(db=db, whitelist=whitelist, blacklist=blacklist)
        # calculate overall rates
        good_stats = [result for result in stats if type(result) is FileStatus]
        branch_rates = [result.branch_rate for result in good_stats
                            if type(result.branch_rate) is Rate]
        line_rates = [result.line_rate for result in good_stats
                            if type(result.line_rate) is Rate]
        line_rate = None
        if line_rates:
            line_rate = reduce(lambda result, rate: result | rate,
                                    line_rates[1:], line_rates[0])
        branch_rate = None
        if branch_rates:
            branch_rate = reduce(lambda result, rate: result | rate,
                                    branch_rates[1:], branch_rates[0])
        # create module xml nodes
        import time
        import lxml.etree as ET
        from lxml.builder import E
        packages = []
        for result in good_stats:
            line_elements = []
            for line in result.lines:
                line_element = E.line(number=str(line), hits=line in result.hit_count and \
                                    str(result.hit_count[line]) or "0")
                if line in result.branches:
                    line_element.set('branch', 'true')
                    line_element.set('condition-coverage', "{0:}% ({1:})".format(
                                        int(result.conditions[line]*100), result.conditions[line]))
                line_elements.append(line_element)
            class_element = E("class", E.methods(), E.lines(*line_elements), name=os.path.basename(result.filename),
                                complexity='0.00', filename=result.filename)
            class_element.set('branch-rate', "%1.4f" % (result.branch_rate or 0.0))
            class_element.set('line-rate', "%1.4f" % (result.line_rate or 0.0))
            package = E.package(E.classes(class_element), name="", complexity='0.00')
            package.set('branch-rate', "%1.4f" % (result.branch_rate or 0.0))
            package.set('line-rate', "%1.4f" % (result.line_rate or 0.0))
            packages.append(package)
        tree = E.coverage(
            E.packages(*packages),
            timestamp=str(int(time.time()*1000)),
            version='3.7',
            complexity='0.00'
        )
        tree.set('branch-rate', '%1.4f' % (branch_rate or 0.0))
        tree.set('line-rate', '%1.4f' % (line_rate or 0.0))
        tree.set('lines-covered', line_rate and str(line_rate.numerator) or "0")
        tree.set('lines-valid', line_rate and str(line_rate.denominator) or "0")
        tree.set('branches-covered', branch_rate and str(branch_rate.numerator) or "0")
        tree.set('branches-valid', branch_rate and str(branch_rate.denominator) or "0")
        return ET.tostring(tree, pretty_print=True, xml_declaration=True,
                            doctype="<!DOCTYPE coverage SYSTEM 'http://cobertura.sourceforge.net/xml/coverage-04.dtd'>")



if __name__ == '__main__':
    #print_stats()
    print get_stats_xml()



'''monkey patch hacks'''
import conf

def patch_coveragepy(dbhost=conf.DBHOST, dbport=conf.DBPORT, dbname=conf.DBNAME):
    '''monkey patch coveragepy in order to fetch stats from our data store'''
    import coverage

    def raw_data(self, arg_ignored):
        '''moncov patched raw data method to fetch stats from moncov data store'''
        import moncov
        db = moncov.conf.get_db(dbhost=dbhost, dbport=dbport, dbname=dbname)
        if hasattr(self, '_moncov_data_cache'):
            return self._moncov_data_cache

        data = {
            'arcs': {},
            'lines': {}
        }
        for filename in moncov.data.filenames(db):
            data['arcs'][filename] = []
            data['lines'][filename] = []
            for arc in moncov.data.filename_arcs(db, filename):
                data['arcs'][filename].append(moncov.data.arc2tuple(arc))
                data['lines'][filename].append(moncov.data.arc2line(arc))
        self._moncov_data_cache = data
        return self._moncov_data_cache

    coverage.data.CoverageData.raw_data = raw_data




import fractions

class Rate(fractions.Fraction):

    def __new__(cls, numerator=0, denominator=None):
        '''avoid reduction'''
        self = super(Rate, cls).__new__(cls, numerator, denominator)
        if denominator is None:
            return self
        gcd = fractions.gcd(numerator, denominator)
        self._numerator *= gcd
        self._denominator *= gcd
        return self

    def __or__(self, other):
        '''grow the portion and the pie size'''
        return type(self)(self.numerator + other.numerator,
                        self.denominator + other.denominator)

    def __repr__(self):
        return '%s(%r, %r)' % (type(self).__name__, self.numerator, self.denominator)

    def __str__(self):
        return '%s/%s' % (self.numerator, self.denominator)

    @property
    def fraction(self):
        '''return a fraction created out of a rate instance'''
        return Fraction(self.numerator, self.denominator)





"""Raw data collector for Coverage."""

import sys
import threading
import conf
import re
from tracer import PyTracer
import logging

log = logging.getLogger(__name__)

class Collector(object):
    # The stack of active Collectors.  Collectors are added here when started,
    # and popped when stopped.  Collectors on the stack are paused when not
    # the top, and resumed when they become the top again.

    def __init__(self, db=None, dbhost=None, dbport=None, dbname=None,
                whitelist=None, blacklist=None):
        self.tracers = []
        self._trace_class = PyTracer
        if db:
            self.dbhost, self.dbport, self.dbname = conf.get_dbdetails(db)
        else:
            self.dbhost = dbhost or conf.DBHOST
            self.dbport = dbport or conf.DBPORT
            self.dbname = dbname or conf.DBNAME
        self.blacklist = conf.get_blacklist(blacklist)
        self.whitelist = conf.get_whitelist(whitelist)
        log.info('using: %r, %r, %r, %r, %r' % (self.dbhost, self.dbport, self.dbname,
                    [x.pattern for x in self.whitelist], [x.pattern for x in self.blacklist]))


    def __repr__(self):
        return "<Collector at 0x%x>" % id(self)

    def _start_tracer(self):
        """Start a new Tracer object, and store it in self.tracers."""
        tracer = PyTracer(dbhost=self.dbhost, dbport=self.dbport,
                dbname=self.dbname, blacklist=self.blacklist, whitelist=self.whitelist)
        fn = tracer.start()
        self.tracers.append(tracer)
        return fn

    # The trace function has to be set individually on each thread before
    # execution begins.  Ironically, the only support the threading module has
    # for running code before the thread main is the tracing function.  So we
    # install this as a trace function, and the first time it's called, it does
    # the real trace installation.

    def _installation_trace(self, frame_unused, event_unused, arg_unused):
        """Called on new threads, installs the real tracer."""
        # Remove ourselves as the trace function
        sys.settrace(None)
        # Install the real tracer.
        fn = self._start_tracer()
        # Invoke the real trace function with the current event, to be sure
        # not to lose an event.
        if fn:
            fn = fn(frame_unused, event_unused, arg_unused)
        # Return the new trace function to continue tracing in this scope.
        return fn

    def start(self):
        """Start collecting trace information."""

        # Check to see whether we had a fullcoverage tracer installed.
        traces0 = []
        if hasattr(sys, "gettrace"):
            fn0 = sys.gettrace()
            if fn0:
                tracer0 = getattr(fn0, '__self__', None)
                if tracer0:
                    traces0 = getattr(tracer0, 'traces', [])

        # Install the tracer on this thread.
        fn = self._start_tracer()

        for args in traces0:
            (frame, event, arg), lineno = args
            try:
                fn(frame, event, arg, lineno=lineno)
            except TypeError:
                raise Exception(
                    "fullcoverage must be run with the C trace function."
                )

        # Install our installation tracer in threading, to jump start other
        # threads.
        threading.settrace(self._installation_trace)



'''data manipulation stuff'''

def arc2tuple(arc):
    '''convert moncov arc "1,2" to a lineno tuple (1, 2)'''
    first, last = arc.split(',')
    return int(first), int(last)

def arc2line(arc):
    '''convert moncov arc "1,2" to an executed lineno 2'''
    first, last = arc.split(',')
    last = int(last)
    if last < 0:
        last = - last
    return last

def filenames(db):
    '''return collected/measured filenames'''
    return db.smembers('filenames')

def filename_arcs(db, filename):
    '''return collected filename arcs'''
    return db.zrange(filename, 0, -1)

def filename_arc_hits(db, filename, arc):
    '''return hit count of a filename arc'''
    return db.zrank(filename, arc)



'''raw data tracer for Coverage'''
import sys
import threading
import logging
log = logging.getLogger(__name__)

_SHOULD_TRACE = {}

class PyTracer(object):

    def __init__(self, dbhost="localhost", dbport=6379, dbname=0, blacklist=[], whitelist=[]):
        self.blacklist = blacklist
        self.whitelist = whitelist
        self._should_process = getattr(sys.modules[__name__], '_SHOULD_TRACE')
        self.last_lineno = -1
        self.last_exc_back = None
        # contains a history of self.last_linenos
        self.stack = []
        try:
            import conf
            self.db = conf.get_db(dbhost, dbport, dbname)
            # create a redis pipeline. pipeline.execute() need not
            # be transactional --- all operations should be just increments
            self.pipeline = self.db.pipeline(transaction=False)
        except Exception as e:
            log.error("got: %r, disabling." % e)
            self.stop()
        else:
            self.enabled = True
            log.info('%r using: %r' % (self, self.pipeline))


    def _trace(self, frame, event, arg_unused):
        """The trace function passed to sys.settrace."""
        # reentrance not allowed
        if not self.enabled:
            return self._trace

        # does the event need processing?
        if event != 'line' and event != 'call' and event != 'return' and event != 'exception':
            return self._trace

        # fetch filename and line number from frame details
        filename, lineno = frame.f_code.co_filename, frame.f_lineno

        # does the filename need processing?
        # the result of blacklist/whitelist filtering is cached here
        # the cache is shared between all tracers of a process
        # the items in the cache stabilize
        # filename, if first encountered, is marked as being traced in redis, too
        # all (even remote) tracers have same white/black list
        # marking of the filenames will happen multiple times but we don't mind
        if filename not in self._should_process:
            # mask-out tracing
            self.enabled = False
            self._should_process[filename] = \
                any([regexp.match(filename) for regexp in self.whitelist]) and \
                not any([regexp.match(filename) for regexp in self.blacklist])
            if self._should_process[filename]:
                try:
                    # mark this filename being processed
                    # this might preempt self.pipeline from time to time
                    # we should be safe
                    self.db.sadd('filenames', filename)
                except Exception as e:
                    log and log.warning('%s._trace: marking %s traced got exception: %r' %  \
                        (self, filename, e))
            # un-mask tracing
            self.enabled = True
        if not self._should_process[filename]:
            # the filename doesn't need processing --- filtered out
            return self._trace

        if self.last_exc_back:
            # handling exc_back situation
            if frame == self.last_exc_back:
                # handling exception
                try:
                    # mask-out tracing
                    self.enabled = False
                    # record the line being executed
                    self.pipeline.zincrby(filename, str(self.last_lineno) + ',' + str(-self.last_exc_back.f_code.co_firstlineno), 1)
                    # flush what we've got and start catching events into pipeline again
                    # in case not all operations made it to the server, print a warning
                    if not all(self.pipeline.execute()):
                        log and log.warning('%s._trace: pipeline.execute missed events' % self)
                except Exception as e:
                    log and log.warning('%s._trace got error: %r' % (self, e))
                finally:
                    # un-mask tracing
                    self.enabled = True
                    self.last_lineno = self.stack.pop()
            self.last_exc_back = None


        # process the event
        if event == 'call':
            self.stack.append(self.last_lineno)
            self.last_lineno = -1

        elif event == 'line':
            try:
                # mask-out tracing
                self.enabled = False
                # record the line being executed
                self.pipeline.zincrby(filename, str(self.last_lineno) + ',' + str(lineno), 1)
            except Exception as e:
                log and log.warning('%s._trace got error: %r' % (self, e))
            finally:
                # un-mask tracing
                self.enabled = True
                self.last_lineno = lineno

        elif event == 'return':
            try:
                # mask-out tracing
                self.enabled = False
                # record return arc
                self.pipeline.zincrby(filename, str(self.last_lineno) + ',' + str(-frame.f_code.co_firstlineno), 1)
                # flush what we've got and start catching events into pipeline again
                # in case not all operations made it to the server, print a warning
                if not all(self.pipeline.execute()):
                    log and log.warning('%s._trace: pipeline.execute missed events' % self)
            except Exception as e:
                # self.__dell__ messes with tracing&logging
                log and log.warning('%s._trace got error: %r' % (self, e))
            finally:
                # un-mask tracing
                self.enabled = True
                self.last_lineno = self.stack.pop()

        elif event == 'exception':
            self.last_exc_back = frame.f_back

        # end processing
        return self._trace

    def start(self):
        """Start this Tracer."""
        self.enabled = True
        sys.settrace(self._trace)
        return self._trace

    def stop(self):
        """Stop this Tracer."""
        self.enabled = False
        sys.settrace(None)

    def get_stats(self):
        """Return a dictionary of statistics, or None."""
        return None

    def __del__(self):
        import sys
        sys.settrace(None)





from rate import Rate
import collections


Line = collections.namedtuple('Line', ['lineno'])
IfLine = collections.namedtuple('IfLine', ['lineno'])
MethodLine = collections.namedtuple('MethodLine', ['lineno'])
ClassLine = collections.namedtuple('ClassLine', ['lineno'])
ModuleLine = collections.namedtuple('ModuleLine', ['lineno'])

class Status(object):
    '''line status object'''
    lineno = None
    linetype = None
    lines = None
    line_rate = None
    branch_rate = None
    hits = None
    linetype = None
    def __init__(self, lineno=0, lines=set(), hits=set(), line_rate=Rate(0, 0),
            branch_rate=Rate(0, 0), linetype=ModuleLine):
        self.linetype = linetype
        self.lineno = lineno
        self.lines = lines
        self.line_rate = line_rate
        self.branch_rate = branch_rate
        self.hits = hits

      def merge(self, other):
        # merge statuses <<=
        assert type(self) is type(other)
        self.lines |= other.lines
        self.lines.add(other.line)
        self.hits |= other.hits
        self.branch_rate |= other.branch_rate
        self.line_rate |= other.line_rate

    @property
    def line(self):
        return self.linetype(lineno=self.lineno)

    def add_line(self, lineno):
        '''add another line and update line rate'''
        line = self.linetype(lineno)
        if line in self.lines:
            # already counted
            return
        self.lines.add(line)
        if lineno in self.hits:
            # executed line
            self.line_rate |= Rate(1, 1)
        else:
            self.line_rate |= Rate(0, 1)

    def __repr__(self):
        return '%r(lineno=%r, lines=%r, hits=%r, line_rate=%r, branch_rate=%r)' % (
                type(self).__name__, self.lineno, self.lines, self.hits, self.line_rate,
                self.branch_rate)

    def __str__(self):
        return 'Status: line_rate = %s, branch_rate = %s' % (self.line_rate, self.branch_rate)



#!/usr/bin/python
import ast
import collections
import sys
import fractions
import redis
import time

#to generate xml
from lxml import etree
from lxml.builder import E
from lxml.builder import ElementMaker

# *_[EX] events: Enter eXit
M_E = 'M_E' # Module
M_X = 'M_X'
F_E = 'F_E' # Function
F_X = 'F_X'
C_E = 'C_E' # Class
C_X = 'C_X'
O_E = 'O_E' # Other
O_X = 'O_X'
IFX = 'IFX' # leaving if
FOX = 'FOX' # leaving for
WHX = 'WHX' # leaving while


# 'States'
BTT = 'BTT'  # stack bottom
CTT = 'CTT'  # bottom of regular class
FTT = 'FTT'  # bottom of regular method/function
IFT = 'IFT'  # if.test attribute
IFB = 'IFB'  # if.body attribute
IFO = 'IFO'  # if.orelse attribute
FOT = 'FOT'  # for.test attribute
FOB = 'FOB'  # for.body attribute
WHT = 'WHT'  # while.test attribute
WHB = 'WHB'  # while.body attribute

# stack actions
NOOP = lambda: lambda stack: stack
PUSH = lambda items: lambda stack: stack + items

# Transition table for cobertura.dtd
TBL = {
        (M_E, BTT): PUSH([BTT]),
        (M_X, BTT): PUSH([BTT]),

        (F_E, BTT): PUSH([BTT, F_E]), # module-level function
        (F_E, CTT): PUSH([CTT, FTT]), # method
        (F_E, FTT): PUSH([FTT, F_E]), # function within method
        (F_E, F_E): PUSH([F_E, F_E]), # function within function
        (F_E, C_E): PUSH([C_E, F_E]), # function within nested class
        (F_E, O_E): PUSH([O_E, F_E]), # function within other statement
        (F_E, IFB): PUSH([IFB, F_E]), # entering function definition within if.body
        (F_E, IFO): PUSH([IFO, F_E]), # entering function definition within if.orelse
        (F_E, FOB): PUSH([FOB, F_E]), # entering function definition within for.body
        (F_E, WHB): PUSH([WHB, F_E]), # entering function definition within while.body
        (F_X, FTT): NOOP(),           # leaving method
        (F_X, F_E): NOOP(),           # leaving any nested function

        (C_E, BTT): PUSH([BTT, CTT]), # entering regular class
        (C_E, CTT): PUSH([CTT, C_E]), # entering class nested in regular class
        (C_E, FTT): PUSH([FTT, C_E]), # entering class within a method
        (C_E, C_E): PUSH([C_E, C_E]), # entering class within a class
        (C_E, F_E): PUSH([F_E, C_E]), # entering class within a function
        (C_E, O_E): PUSH([O_E, C_E]), # entering class within other statement
        (C_X, CTT): NOOP(),           # leaving regular class
        (C_X, C_E): NOOP(),           # leaving any nested function
        (C_E, IFB): PUSH([IFB, C_E]), # entering class definition within if.body
        (C_E, FOB): PUSH([FOB, C_E]), # entering class definition within for.body
        (C_E, WHB): PUSH([WHB, C_E]), # entering class definition within while.body
        (C_E, IFO): PUSH([IFO, C_E]), # entering class definition within if.orelse

        (O_E, BTT): PUSH([BTT, O_E]), # entering other module-level statement
        (O_E, FTT): PUSH([FTT, O_E]), # entering other method-level statement
        (O_E, CTT): PUSH([CTT, O_E]), # entering other regular-class-level statement
        (O_E, F_E): PUSH([F_E, O_E]), # entering other (nested-)function-level statement
        (O_E, C_E): PUSH([C_E, O_E]), # entering other nested-class-level statement
        (O_E, O_E): PUSH([O_E, O_E]), # entering other (nested) statemetn
        (O_E, IFT): PUSH([IFT, O_E]), # entering other statement within if.test
        (O_E, IFB): PUSH([IFB, O_E]), # entering other statement within if.body

        (O_E, FOT): PUSH([FOT, O_E]), # entering other statement within for.test
        (O_E, FOB): PUSH([FOB, O_E]), # entering other statement within for.body

        (O_E, WHT): PUSH([WHT, O_E]), # entering other statement within while.test
        (O_E, WHB): PUSH([WHB, O_E]), # entering other statement within while.body

        (O_E, IFO): PUSH([IFO, O_E]), # entering other statement within if.orelse
        (O_X, O_E): NOOP(),           # leaving other (nested) statement
        (O_X, FTT): PUSH([FTT]),
        (O_X, CTT): PUSH([CTT]),
        (O_X, BTT): PUSH([BTT]),

        (IFT, BTT): PUSH([BTT, IFT]), # entering if statement within a module line
        (IFT, FTT): PUSH([FTT, IFT]), # entering if statement within a regular method line
        (IFT, CTT): PUSH([CTT, IFT]), # entering if statement within a regular class line
        (IFT, C_E): PUSH([C_E, IFT]), # entering if statement within a nested class line
        (IFT, F_E): PUSH([F_E, IFT]), # entering if statement within a nested function line
        (IFT, O_E): PUSH([O_E, IFT]), # entering if statement within a (sub-)statement line
        (IFT, IFB): PUSH([IFB, IFT]), # entering nested if statement from within if.body

        (IFT, FOB): PUSH([FOB, IFT]), # entering nested if statement from within for.body
        (IFT, WHB): PUSH([WHB, IFT]), # entering nested if statement from within while.body

        (IFT, IFO): PUSH([IFO, IFT]), # entering nested if statement from within if.orelse
        (IFB, IFT): PUSH([IFB]),      # entering if.body
        (IFO, IFB): PUSH([IFO]),      # entering if.orelse
        (IFX, IFB): NOOP(),           # leaving if statement from within if.body (no orelse present)
        (IFX, IFO): NOOP(),           # leaving if statement from within of.orelse # mistake?
        (IFX, FOB): NOOP(),           # leaving if statement from within for.body
        (IFX, WHB): NOOP(),           # leaving if statement from within while.body
        (IFX, IFO): NOOP(),           # leaving if statement from within if.orelse



        (FOT, BTT): PUSH([BTT, FOT]), # entering for statement within a module line
        (FOT, FTT): PUSH([FTT, FOT]), # entering for statement within a regular method line
        (FOT, CTT): PUSH([CTT, FOT]), # entering for statement within a regular class line
        (FOT, C_E): PUSH([C_E, FOT]), # entering for statement within a nested class line
        (FOT, F_E): PUSH([F_E, FOT]), # entering for statement within a nested function line
        (FOT, O_E): PUSH([O_E, FOT]), # entering for statement within a (sub-)statement line
        (FOT, FOB): PUSH([FOB, FOT]), # entering nested for statement from within for.body
        (FOT, IFB): PUSH([IFB, FOT]), # entering nested for statement from within if.body
        (FOT, WHB): PUSH([WHB, FOT]), # entering nested for statement from within while.body
        (FOT, IFO): PUSH([IFO, FOT]), # entering nested for statement from within if.orelse
        (FOB, FOT): PUSH([FOB]),      # entering for.body
        (FOX, FOB): NOOP(),           # leaving for statement from within for.body
        (FOX, IFB): NOOP(),           # leaving for statement from within if.body
        (FOX, IFO): NOOP(),           # leaving for statement from within if.orelse
        (FOX, WHB): NOOP(),           # leaving for statement from within while.body


        (WHT, BTT): PUSH([BTT, WHT]), # entering while statement within a module line
        (WHT, FTT): PUSH([FTT, WHT]), # entering while statement within a regular method line
        (WHT, CTT): PUSH([CTT, WHT]), # entering while statement within a regular class line
        (WHT, C_E): PUSH([C_E, WHT]), # entering while statement within a nested class line
        (WHT, F_E): PUSH([F_E, WHT]), # entering while statement within a nested function line
        (WHT, O_E): PUSH([O_E, WHT]), # entering while statement within a (sub-)statement line
        (WHT, FOB): PUSH([FOB, WHT]), # entering nested while statement from within for.body
        (WHT, IFB): PUSH([IFB, WHT]), # entering nested while statement from within if.body
        (WHT, WHB): PUSH([WHB, WHT]), # entering nested while statement from within while.body
        (WHT, IFO): PUSH([IFO, WHT]), # entering nested while statement from within if.orelse
        (WHB, FOT): PUSH([FOB]),      # entering while.body
        (WHB, WHT): PUSH([WHB]),      # entering while.body from within while.test
        (WHX, FOB): NOOP(),           # leaving while statement from within for.body
        (WHX, IFB): NOOP(),           # leaving while statement from within if.body
        (WHX, IFO): NOOP(),           # leaving while statement from within if.orelse
        (WHX, WHB): NOOP(),           # leaving while statement from within while.body
        (WHX, WHT): NOOP(),           # leaving while statement from within while.body


}


#get data from mongo

#create xml skeleton
E = ElementMaker()
xml = (E.coverage(E.sources(E.source(),E.source), E.packages()))
xml_tree = etree.ElementTree(xml)
xml.insert(0, etree.Comment("Generated by pushdown_moncov.py: http://github.com/python-moncov/bla-bla"))



class PushDownAutomaton(object):

    # by default, use `lambda event, stack_event: None' callbacks for any event
    def __init__(self, stack=None, callbacks=collections.defaultdict(lambda: lambda event, stack_event: None)):
        if stack is None:
            stack = [BTT]
        self.reset(stack)
        self.callbacks = callbacks

    def reset(self, stack=None):
        if stack is None:
            stack = [BTT]
        self.stack = stack

    def __repr__(self):
        return type(self).__name__ + '(stack=%r)'% self.stack

    def __call__(self, event):
        stack_event = self.stack.pop()
        # perform callback
        self.callbacks[(event, stack_event)](event, stack_event)
        # perform stack manipulation based on the transition table
        self.stack = TBL[(event, stack_event)](self.stack)
        #import pprint
        #print pprint.pformat(self.stack), event


class Stack(object):
    def __init__(self, data=[]):
        self.data = data

    @property
    def top(self):
        return self.data[-1]

    @top.setter
    def top(self, other):
        self.data[-1] = other

    @property
    def depth(self):
        return len(self.data)

    def push(self, item):
        self.data.append(item)

    def pop(self):
        return self.data.pop()


class Rate(fractions.Fraction):

    def __new__(cls, numerator=0, denominator=None):
        '''avoid reduction'''
        self = super(Rate, cls).__new__(cls, numerator, denominator)
        if denominator is None:
            return self
        gcd = fractions.gcd(numerator, denominator)
        self._numerator *= gcd
        self._denominator *= gcd
        return self

    def __or__(self, other):
        '''grow the portion and the pie size'''
        return type(self)(self.numerator + other.numerator,
                        self.denominator + other.denominator)

    def __and__(self, other):
        '''grow the portion and pie size if both self and other are not zero'''
        if self != 0 and other != 0:
            return self | other
        elif self == 0 and other != 0:
            return other
        elif self !=0 and other == 0:
            return self
        else:
            return Rate(0)


    def __repr__(self):
        return '%s(%r, %r)' % (type(self).__name__, self.numerator, self.denominator)

    @classmethod
    def as_fraction(cls, rate):
        '''return a fraction created out of a rate instance'''
        return fractions.Fraction(rate.numerator, rate.denominator)


Line = collections.namedtuple('Line', ['lineno'])
IfLine = collections.namedtuple('IfLine', ['lineno'])
ForLine = collections.namedtuple('ForLine', ['lineno'])
WhileLine = collections.namedtuple('WhileLine', ['lineno'])
MethodLine = collections.namedtuple('MethodLine', ['lineno'])
ClassLine = collections.namedtuple('ClassLine', ['lineno'])
ModuleLine = collections.namedtuple('ModuleLine', ['lineno'])

class Status(object):
    lineno = None
    linetype = None
    lines = None
    branch_rate = None
    linetype = None
    place = None

    def __init__(self, lineno=0, lines=set(),
            branch_rate=Rate(0), linetype=ModuleLine, hit_count = None, place = None, class_data = None):
        self.linetype = linetype
        self.lineno = lineno
        self.lines = lines
        self.branch_rate = branch_rate
        self.hit_count = hit_count or None
        self.place = place or [0,0]
        self.class_data = class_data or [0,0, 0, 0, 0, 0]

        # class_data[0] = counter for methods hits
        # class_data[1] = counter for methods lines
        # class_data[2] = counter for class hits
        # class_data[3] = counter for class lines
        # class_data[4] = counter for package hits
        # class_data[5] = counter for paclage lines

    @property
    def lines(self):
        return self._lines

    @lines.setter
    def lines(self, other_set):
        # ensure self.lineno is kept in the set
        self._lines = other_set
        #self._lines.add(self.line)

    def merge(self, other):
        # merge statuses <<=
        # assert type(self) is type(other)
        self.lines |= other.lines
        self.branch_rate &= other.branch_rate
        self.place = other.place
        self.class_data = other.class_data



    @property
    def line(self):
        return self.linetype(lineno=self.lineno)

    def add_line(self, lineno):
        '''add another line and update line rate'''

        line = self.linetype(lineno)

        if line not in self.lines:

            if lineno in self.hit_count.keys():
                self.class_data[4] += 1
                hits_ = str(int(self.hit_count[lineno]))
            else:
                hits_ = '0'

            self.class_data[5] += 1

            if self.linetype == ClassLine:
                try:
                    self.xml_status = xml[-1][-1][-1][-1][-1]
                    self.xml_status.append(E(tag='line', hits = hits_, number=str(lineno), branch='false'))
                except:
                    pass

                if lineno in self.hit_count.keys():
                    self.class_data[2] += 1
                self.class_data[3] +=1


            if self.linetype == MethodLine:
                try:
                    self.xml_status = xml[-1][-1][-1][-1][0][-1][-1]
                    self.xml_status.append(E(tag='line', hits = hits_, number=str(lineno), branch='false'))
                except:
                    pass

                if lineno in self.hit_count.keys():
                    self.class_data[0] += 1
                    self.class_data[2] += 1

                self.class_data[1] += 1
                self.class_data[3] += 1


            if self.linetype == IfLine or self.linetype == WhileLine or self.linetype == ForLine:

                if self.place[0] == 'class':

                    if lineno in self.hit_count.keys():
                        self.class_data[2] += 1
                    self.class_data[3] += 1

                    if self.place[1] == 1:

                        try:
                            self.xml_status = xml[-1][-1][-1][-1][-1]
                            self.xml_status.append(E.line(E.conditions(E.condition(number="0", type="jump", coverage="100%")), hits = hits_, number=str(lineno), branch='true'))


                            if self.branch_rate.numerator == self.branch_rate.denominator:
                                branch_rate = "100% (2/2)"
                                xml[-1][-1][-1][-1][-1][-1][-1][-1].set('coverage', "100%")
                            if self.branch_rate.denominator !=0 :
                                branch_rate = str(self.branch_rate.numerator*100/self.branch_rate.denominator) + "% (" + str(self.branch_rate) + ")"
                                xml[-1][-1][-1][-1][-1][-1][-1][-1].set('coverage', str(self.branch_rate.numerator*100/self.branch_rate.denominator)+"%")

                            xml[-1][-1][-1][-1][-1][-1].set("condition-coverage", branch_rate)

                        except:
                            pass

                    else:
                        try:
                            self.xml_status = xml[-1][-1][-1][-1][-1]
                            self.xml_status.append(E(tag='line', hits = hits_, number=str(lineno), branch='false'))
                        except:
                            pass


                if self.place[0] == 'method':

                    if lineno in self.hit_count.keys():
                        self.class_data[0] += 1
                        self.class_data[2] += 1

                    self.class_data[1] += 1
                    self.class_data[3] += 1

                    if self.place[1] == 1:


                        try:
                            self.xml_status = xml[-1][-1][-1][-1][0][-1][-1]
                            self.xml_status.append(E.line(E.conditions(E.condition(number="0", type="jump")), hits = hits_, number=str(lineno), branch='true'))

                            if self.branch_rate.numerator == self.branch_rate.denominator:
                                branch_rate = "100% (2/2)"
                                xml[-1][-1][-1][-1][0][-1][-1][-1][-1][-1].set('coverage', "100%")
                            else:
                                if self.branch_rate.denominator !=0 :
                                    branch_rate = str(self.branch_rate.numerator*100/self.branch_rate.denominator) + "% (" + str(self.branch_rate) + ")"
                                    xml[-1][-1][-1][-1][0][-1][-1][-1][-1][-1].set('coverage', str(self.branch_rate.numerator*100/self.branch_rate.denominator)+"%")

                            xml[-1][-1][-1][-1][0][-1][-1][-1].set("condition-coverage", branch_rate)
                        except:
                            pass
                    else:
                        try:
                            self.xml_status = xml[-1][-1][-1][-1][0][-1][-1]
                            self.xml_status.append(E(tag='line', hits = hits_, number=str(lineno), branch='false'))
                        except:
                            pass


        if line in self.lines:
            # already counted
            return

        self.lines.add(line)


    def __repr__(self):
        return '%r(lineno=%r, lines=%r, branch_rate=%r, place = %r, class_data=%r)' % (
                type(self), self.lineno, self.lines,
                self.branch_rate, self.place, self.class_data)

    def __str__(self):
        return 'Status: branch_rate = %s' % (self.branch_rate)


class Visitor(ast.NodeVisitor):

    def __init__(self, hit_count={}, filename = None, place = None, class_data = None):
        callbacks = {
            (M_E, BTT): self.init_module,
            (M_X, BTT): self.exit_module,
            (C_E, BTT): self.init_class,
            (F_E, CTT): self.init_method,
            (C_X, CTT): self.exit_nested,
            (F_X, FTT): self.exit_nested,


            (IFT, BTT): self.init_if,
            (IFT, CTT): self.init_if,
            (IFT, FTT): self.init_if,
            (IFT, F_E): self.init_if,
            (IFT, C_E): self.init_if,
            (IFT, O_E): self.init_if,
            (IFT, IFT): self.init_if,
            (IFT, IFB): self.init_if,
            (IFT, IFO): self.init_if,
            (IFT, WHB): self.init_if,
            (IFT, FOB): self.init_if,



            (FOT, BTT): self.init_for,
            (FOT, CTT): self.init_for,
            (FOT, FTT): self.init_for,
            (FOT, F_E): self.init_for,
            (FOT, C_E): self.init_for,
            (FOT, O_E): self.init_for,
            (FOT, IFT): self.init_for,
            (FOT, IFB): self.init_for,
            (FOT, IFO): self.init_for,
            (FOT, FOT): self.init_for,
            (FOT, FOB): self.init_for,
            (FOT, WHT): self.init_for,
            (FOT, WHB): self.init_for,


            (WHT, BTT): self.init_while,
            (WHT, CTT): self.init_while,
            (WHT, FTT): self.init_while,
            (WHT, F_E): self.init_while,
            (WHT, C_E): self.init_while,
            (WHT, O_E): self.init_while,
            (WHT, IFT): self.init_while,
            (WHT, IFB): self.init_while,
            (WHT, IFO): self.init_while,
            (WHT, FOT): self.init_while,
            (WHT, FOB): self.init_while,
            (WHT, WHT): self.init_while,
            (WHT, WHB): self.init_while,


            (IFX, IFB): self.exit_nested,
            (IFX, IFO): self.exit_nested,
            (IFX, FOB): self.exit_nested,
            (IFX, WHB): self.exit_nested,



            (FOX, IFB): self.exit_nested,
            (FOX, IFO): self.exit_nested,
            (FOX, FOB): self.exit_nested,
            (FOX, WHB): self.exit_nested,

            (WHX, IFB): self.exit_nested,
            (WHX, IFO): self.exit_nested,
            (WHX, FOB): self.exit_nested,
            (WHX, WHB): self.exit_nested,

        }
        self.callbacks = collections.defaultdict( \
            lambda: self.line_handler,
            callbacks.items()
        )
        self.pda = PushDownAutomaton(callbacks=self.callbacks)
        self.node = None
        # hit_count is a mapping: lineno->line_hits
        self.hit_count = hit_count

        # set of lines explored during parsing
        self.lines = set()

        self.report = None
        self.filename = filename or None
        self.place = place or [0, 0]
        self.xml_status = xml[-1]

        self.class_data = class_data or [0, 0, 0, 0, 0, 0]


        super(Visitor, self).__init__()

    def visit_Module(self, node):
        self.place[0] = "module"
        self.generic_visit(node, (M_E, M_X))

    def visit_FunctionDef(self, node):
        self.generic_visit(node, (F_E, F_X))

    def visit_ClassDef(self, node):
        self.generic_visit(node, (C_E, C_X))

    def visit_If(self, node):
        # need to distinguish between test/body/orelse if statement subtrees
        self.node = node
        self.pda(IFT)
        self.visit(node.test)
        self.place[1] = 0
        self.pda(IFB)
        for sub_node in node.body:
            self.visit(sub_node)
        if node.orelse:
            self.pda(IFO)
        for sub_node in node.orelse:
            self.visit(sub_node)
        self.pda(IFX)


    def visit_For(self, node):
        self.node = node
        self.pda(FOT)
        self.visit(node.target)
        self.visit(node.iter)
        self.place[1] = 0
        self.pda(FOB)
        for subnode in node.body + node.orelse:
            self.visit(subnode)
        #self.add_branch(node)
        self.pda(FOX)

    def visit_While(self, node):
        self.node = node
        self.pda(WHT)
        self.visit(node.test)
        self.place[1] = 0
        self.pda(WHB)
        for subnode in node.body + node.orelse:
            self.visit(subnode)
        #self.add_branch(node)
        self.pda(WHX)


    def generic_visit(self, node, events=(O_E, O_X)):
        self.node = node
        #print "::%s" % type(self.node)
        self.pda(events[0])
        super(Visitor, self).generic_visit(node)
        self.pda(events[1])

    def line_handler(self, event, stack_event):
        if hasattr(self.node, 'lineno'):
            self.report.top.add_line(self.node.lineno)



    def init_module(self, event, stack_event):
        self.place[0] = 'module'
        self.report = Stack([Status(lineno=1, linetype=ModuleLine, hit_count = self.hit_count, place = self.place, class_data = self.class_data)])
        self.xml_status.append(E(tag = "package", name = self.filename))
        self.xml_status = xml[-1][-1]
        self.xml_status.append(E(tag = "classes"))

    def exit_module(self, event, stack_event):
        self.xml_status = xml[-1][-1]
        if self.class_data[5] != 0:
            module_line_rate = float(self.class_data[4])/self.class_data[5]
        else:
            module_line_rate = 0
        xml[-1][-1].set("line-rate", str(module_line_rate))
        if self.report.top.branch_rate.numerator != 0:
            branch_rate = float(self.report.top.branch_rate.numerator)/self.report.top.branch_rate.denominator
        else:
            branch_rate = 1

        xml[-1][-1].set("branch-rate", str(branch_rate))
        xml[-1][-1].set("complexity", str("to_be_counted"))

        pass

    def init_class(self, event, stack_event):
        self.place[0] = "class"
        self.report.push(Status(lineno=self.node.lineno, linetype=ClassLine, hit_count = self.hit_count, place=self.place, class_data = self.class_data))
        self.xml_status = xml[-1][-1][-1]
        self.xml_status.append(E(tag = "class", name = self.node.name, filename =self.filename))
        self.xml_status = xml[-1][-1][-1][-1]
        self.xml_status.append(E(tag = "methods"))
        self.xml_status.append(E(tag = "lines"))


    def exit_nested(self, event, stack_event):

        if event == C_X:
            self.place[0] = "module"
            if self.class_data[3] != 0:
                class_line_rate = float(self.class_data[2])/self.class_data[3]
            else:
                class_line_rate = 0
            xml[-1][-1][-1][-1].set("line-rate", str(class_line_rate))

            if self.report.top.branch_rate.numerator != 0:
                branch_rate = float(self.report.top.branch_rate.numerator)/self.report.top.branch_rate.denominator
            else:
                branch_rate = 1

            xml[-1][-1][-1][-1].set("branch-rate", str(branch_rate))
            xml[-1][-1][-1][-1].set('complexity', 'to_be_counted')

            self.class_data[2] = 0
            self.class_data[3] = 0



        if event == F_X:

            if self.class_data[1] != 0:
                method_line_rate = float(self.class_data[0])/self.class_data[1]
            else:
                method_line_rate = 0
            xml[-1][-1][-1][-1][0][-1].set("line-rate", str(method_line_rate))
            #self.class_data = [x if i not in {0,1} else 0 for i, x in enumerate(self.class_data)]

            if self.report.top.branch_rate.numerator != 0:
                branch_rate = float(self.report.top.branch_rate.numerator)/self.report.top.branch_rate.denominator
            else:
                branch_rate = 1
            xml[-1][-1][-1][-1][0][-1].set("branch-rate", str(branch_rate))

            self.class_data[0] = 0
            self.class_data[1] = 0


        status = self.report.pop()
        self.report.top.merge(status)


    def init_method(self, event, stack_event):
        self.place[0] = "method"
        self.report.push(Status(lineno=self.node.lineno, linetype=MethodLine, hit_count = self.hit_count, place = self.place, class_data = self.class_data))

        self.xml_status = xml[-1][-1][-1][-1][0]
        self.xml_status.append(E(tag = "method", name = self.node.name, signature = 'default'))
        self.xml_status = xml[-1][-1][-1][-1][0][-1]
        self.xml_status.append(E(tag = "lines"))
        self.xml_status = xml[-1][-1][-1][-1][0][-1][-1] # new

    def init_if(self, event, stack_event):
        self.place[1] = 1
        self.report.push(Status(lineno=self.node.lineno, linetype=IfLine, hit_count = self.hit_count, branch_rate=Rate(1, 2), place = self.place, class_data = self.class_data))


    def init_for(self, event, stack_event):
        self.place[1] = 1
        self.report.push(Status(lineno=self.node.lineno, linetype=ForLine, hit_count = self.hit_count, branch_rate=Rate(1, 2), place = self.place, class_data = self.class_data))

    def init_while(self, event, stack_event):
        self.place[1] = 1
        self.report.push(Status(lineno=self.node.lineno, linetype=WhileLine, hit_count = self.hit_count, branch_rate=Rate(1, 2), place = self.place, class_data = self.class_data))


    @property
    def status(self):
        '''return top of self.report stack'''
        return self.report.top


###

def generate_xml(dbhost = "localhost", dbport = 6379, dbname = 0, output='moncov_branch.xml'):
    import moncov
    db = moncov.conf.get_db(dbhost, dbport, dbname)

    for filename in moncov.data.filenames(db):

        if filename.startswith('<'):
            continue

        try:
            with open(filename) as fd:
                xml[-2][-2].text = filename
                xml[-2][-1].text = '--source'
                hit_list = {}
                for arc in moncov.data.filename_arcs(db, filename):
                    hit_list[moncov.data.arc2line(arc)] = moncov.data.filename_arc_hits(db, filename, arc)

                src = fd.read()
                print "Opened file: %s" % filename

            tree = ast.parse(src)
            #print src
            # import astpp
            #print astpp.dump(tree, annotate_fields=True, include_attributes=True)
            visitor = Visitor(hit_count = hit_list, filename = filename)
            #visitor.filename = f
            #print ast.dump(tree)
            visitor.visit(tree)

            import pprint
            print pprint.pformat(visitor.status.lines), len(visitor.status.lines)
            print visitor.status

        except IOError as e:
            print "Can't read: %s" % e.message, filename
            #sys.exit(2)


    f = open(output, 'w')
    f.write(etree.tostring(xml_tree, pretty_print=True, xml_declaration=True, doctype="<!--DOCTYPE coverage SYSTEM 'http://cobertura.sourceforge.net/xml/coverage-04.dtd'-->"))
    f.close()

if __name__ == "__main__":
    start_time = time.time()
    generate_xml()
    finish_time = time.time()

    run_time = finish_time - start_time
    print run_time


#tree = ast.parse(src)
#print src
#import astpp
#print astpp.dump(tree, annotate_fields=True, include_attributes=True)



#visitor = Visitor(hit_count={66: 3})
#visitor.visit(tree)






import moncov
import pymongo
import logging
log = logging.getLogger(__name__)

#  => {_id: {file: <path>, line: <lineno>}, value: <hit-count>}
_MAP = '''function(){emit({file: this.file, line: this.line}, 1)}'''

# {_id: {file: <path>, line: <lineno>}, value: <hit-count>}
# sum per hit-counts
_REDUCE = '''function(key, values){return Array.sum(values)}'''


def update(db=None, dbhost=moncov.conf.DBHOST, dbport=moncov.conf.DBPORT,
            dbname=moncov.conf.DBNAME):
    '''map-reduce events to lines collection'''
    if db is None:
        db = moncov.conf.get_db(dbhost=dbhost, dbport=dbport, dbname=dbname)
    # uses 2 collections
    # - lines: indexed by filename,lineno; fields: value: hit-count
    # - events: capped collection for short-term filename,lineno events
    # ensure proper index by _id.file, _id.line
    db.lines.ensure_index([('_id.file', pymongo.ASCENDING), ('_id.line', pymongo.ASCENDING)],
            unique=True, drop_dups=True, sparse=True)
    try:
        pivot = db.last_event.find({}, sort=[('$natural', 1)], limit=1)[0]
    except IndexError as e:
        # init not yet called
        return
 
    # figure out the last used _id to avoid double-counting
    # - the pivot: no records newer than pivot are being counted
    # after the test&set succeeded
    try:
        last_event = db.events.find({}, sort=[('$natural', -1)], limit=1)[0]
    except IndexError as e:
        # no events to count
        return
    original_last = db.last_event.find_and_modify({'event_id': pivot['event_id']},
            update={'event_id': last_event['_id']})
    # limit the counting with map-reduce query based on the last_event._id
    if original_last is None:
        # race detected --- some other process already counted the interval
        return
    else:
        # will limit the map-reduce to id newer than original
        # last.event_id
        query = {'_id': {'$gt': original_last['event_id'], '$lte': last_event['_id']}}
    # update the hit-counts, "merge-back" with reduce
    response = db.events.map_reduce(map=_MAP, reduce=_REDUCE, query=query,
            out={'reduce': 'lines'}, full_response=True)
    log.info(response)
    #import pprint
    #print pprint.pformat(response)

if __name__ == '__main__':
    update()
